;;; Bulldozer - Play
;;;
;;; - Load the map, used Tiled to draw the map
;;;   - it includes some rocks
;;;   - the rocks and tiles have collision boxes
;;;   - there are two layers: tiles and objects
;;;   - use Simple Tiled Implementation to load the map
;;;   - tell is we will use Box2D
;;; - Handle collisions: un-movable map elements, movable rocks,
;;;   movable bulldozer; the physics engine does that
;;; - Use physics
;;;   - create world and sprite (the bulldozer)
;;;   - will need to decide about movement and pushing,
;;;     this may be complicated
;;;   - implement a maximum speed with power left-over for pushing
;;; - Process WASD to move the bulldozer
;;;   - how much does the physics engine do?
;;;   - do I make two tracks and apply a force to them?
;;;     this would handle turning, but WASD might be "interesting"
;;; - Move the map (the world) when the dozer get close to the edge
;;;   off the screen, handle the edge of the map.
;;; - Find a way to transition to another map (I want a big world).
;;;

(local sti (require "lib.sti"))
(local sound (require "sound"))

;; constants and utility functions
;; -------------------------------

; epsilon - used to see if a value is close enough to zero
;
(local epsilon 0.000001)
(local close-dir (/ math.pi 16))

; normalize an angle in radians
;
; I want the range to be - Pi to + Pi
; Some have reposted very large angle values.
;
(fn normalize-rad [a]
  (math.fmod a (* 2 math.pi)))
;(fn normalize-rad [a]
;  (let [n (math.fmod a (* 2 math.pi))]
;    (if (> n math.pi)
;      (- n math.pi)
;      n)))
(fn normalize-body-angular [b]
  (let [angle (b:getAngle)]
    (b:setAngle (normalize-rad angle))))

; vector sum
;
(fn vector-sum [x-1 y-1 x-2 y-2]
  (values (+ x-1 x-2) (+ y-1 y-2)))

; square the number, saves typing and avoids errors
;
(fn sqr [x]
  (* x x))

; sum of squares
;
(fn sum-of-squares [a b]
  (+ (sqr a) (sqr b)))


; I think I need to set this early
;
; the bulldozer is 2.5 to 3.0 meters long and its sprite is
; 65 pixels (along length), so 20 pixels per metre is OK
;
(local pixels-per-meter 20)
(love.physics.setMeter pixels-per-meter)

; convert pixels to world metres
;
(fn pixels-to-meters [n]
  (/ n pixels-per-meter))
(fn meters-to-pixels [n]
  (* n pixels-per-meter))

;; m a p
;; -----

; I made the map with a tool called Tiled flexible level editor,
; it exports to a Lua file; make sure to ask it to embed the image.
; I plan to use the LÖVE2D built-in physics engine, it is based
; on Box2D.
;
; There are two layers; when I dump the layers table, they appear
; twice: once with index numbers and once with keys.
;
;   Object Layer 1 : table: 0x41d5bf98
;   Tile Layer 1   : table: 0x418797b0
;

; Load the generated Lua code. It describes the map using a tile set,
; a single PNG file with many tiles. It includes the collision boxes
; for the tiles and rocks.
;
; The map should have a tile layer for the terrain and an object
; layer for the rocks. I may also place the bulldozer on the map
; (or at least a place holder); the starting position depends on
; the map's layout (and the puzzle to solve).
;
(local map (sti "assets/map-test.lua" ["box2d"]))

; debug : show some things from the map
;
;(print "map")
;(each [key value (pairs map)]
;  (print key value))

; setup physics, the numbers are for horizontal and vertical gravity
; we play top-down, so they are both zero
;
(local world (love.physics.newWorld 0.0 0.0))

; constants for shifting gears
;
(local low-gear 1)   ; slow & powerful
(local high-gear 2)  ; faster & less powerful

; bulldozer
;
; the image I am using is 65 x 41 pixels
; the centre pixel is at 33,21
;
(local dozer {:img (love.graphics.newImage "assets/bulldozer-player.png")
              :x 400  ; I should pull these two
              :y 300  ;   from the map
              :gear low-gear  ; start in first gear
              :max-speed 25  ; maximum in low gear, twice this in high gear
              :max-speed-2 625  ; will need the square
              :accel-force 200000  ; will experiment to find a good value
              :brake-force 700000  ; will experiment to find a good value
              :dir (math.rad 0)  ; the image naturally points right
              :turn-force 2000  ; will experiment to find a good value
              :max-turn-rate (/ math.pi 8)  ; radians per second
              :brake-turn-force 4000  ; will experiment to find a good value
              :scale-x 1.0
              :scale-y 1.0
              :width 65   ; pixels
              :height 41  ; pixels
              :origin-x 33  ; set origin to centre of bulldozer
              :origin-y 21})

; state of the game
;
(local game-state {:flags {} ; for future use
                   ;; for repl debugging
                   :map map
                   :world world
                   :dozer dozer
                   :echo (fn echo [s msg] (table.insert s.messages 1 msg))})

;; so we can access game state through the repl
;;
(global gs game-state)

;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;;

; this will be used in the update callback
;
(fn move-bulldozer [dt]

  ; bulldozer movement
  ;
  ; - the bulldozer effectively has only two speeds: slow &
  ;   powerful and faster & less powerful
  ; - the bulldozer does not maintain speed, a key must be pressed
  ;   for it to move
  ; - function of the keys
  ;     up & down  and W & S move forward and back
  ;   left & right and A & D turn counter-clockwise and clockwise
  ; - keys can (and will) be combined, opposing keys cancel each
  ;   other
  ;

  ; change direction
  ;
  (if

    ; turn right
    ;
    ; apply negative torque up to maximum angular velocity
    ;
    (love.keyboard.isDown "right" "d")

      (if (not (love.keyboard.isDown "left" "a"))
        ; turn right - angular velocity is in radians per second
        (let [ turn-rate (dozer.body:getAngularVelocity) ]
;         (print "turn right")
          ; positive rad is right (clockwise), use positive max-turn-rate
          (if (<= (- turn-rate epsilon) dozer.max-turn-rate)
            ; close to max, set to max and remove torque
            (do
              (dozer.body:setAngularVelocity dozer.max-turn-rate)
              (dozer.body:applyTorque 0.0))
            (dozer.body:applyTorque dozer.turn-force))))

    ; turn left
    ;
    ; apply positive torque up to maximum angular velocity
    ;
    (love.keyboard.isDown "left" "a")

      (if (not (love.keyboard.isDown "right" "d"))
        ; turn left - angular velocity is in radians per second
        (let [ turn-rate (dozer.body:getAngularVelocity) ]
;         (print "turn left")
          ; negative rad is left (counterclockwise), use negative max-turn-rate
          (if (>= (+ turn-rate epsilon) (- dozer.max-turn-rate))
            ; close to max, set to max and remove torque
            (do
              (dozer.body:setAngularVelocity (- dozer.max-turn-rate))
              (dozer.body:applyTorque 0.0))
            (dozer.body:applyTorque (- dozer.turn-force)))))

    ; else
    ;
    ; quickly kill any angular velocity
    ;
    (let [turn-rate (dozer.body:getAngularVelocity)]
;     (print "no turn - kill rotation")
      (if
        ; close to zero, set to zero and remove torque
        (< (math.abs turn-rate) epsilon)
          (do
            (dozer.body:setAngularVelocity 0.0)
            (dozer.body:applyTorque 0.0))
        (< turn-rate 0.0)  ; negative - turning right
          (dozer.body:applyTorque dozer.brake-turn-force)
        ; else:  positive - turning left
          (dozer.body:applyTorque (- dozer.brake-turn-force)))))

  ; move
  ;
  ; if I apply the force in the current direction,
  ; it will not be much different from applying it
  ; in the new direction (after turning, if any)
  ;
  ; I assume that I will always be moving in either
  ; - the same direction as I am pointing, or
  ; - the opposite direction as I am pointing
  ; - a bulldozer does not move sideways
  ; - give the test some latitude
  ;
  (let [point-dir (dozer.body:getAngle)  ; the bulldozer is pointing this way
        (speed-x speed-y) (dozer.body:getLinearVelocity)  ; and moving at this speed
        move-dir (math.atan2 speed-y speed-x)  ; and moving this way
        max-speed   (* dozer.max-speed dozer.gear)]

;   (print "speed x  " speed-x)
;   (print "speed y  " speed-y)
;   (print "pointing " point-dir)
;   (print "moving   " move-dir)

    (if

      ; move forward
      ;
      ; apply a force to come-up to maximum speed
      ;
      ; looks like I have to deal with direction
      ;
      (love.keyboard.isDown "up" "w")

        (if (not (love.keyboard.isDown "down" "s"))
          ; move forward
          (let [max-speed-x (* (math.cos point-dir) max-speed)
                max-speed-y (* (math.sin point-dir) max-speed)]
;           (print "move forward")
;           (print "max speed x " max-speed-x)
;           (print "max speed y " max-speed-y)
            ; compare magnitude and direction
            ;   forward is small direction difference
            ;
            (if (and (< (+ (sum-of-squares speed-x speed-y) epsilon) dozer.max-speed-2)
                     (< (math.abs (normalize-rad (- point-dir move-dir))) close-dir))
              ; still accelerating
              (let [force-x (* (math.cos point-dir)
                               (/ dozer.accel-force dozer.gear))
                    force-y (* (math.sin point-dir)
                               (/ dozer.accel-force dozer.gear))]
;               (print "accelerating")
                (dozer.body:applyForce force-x force-y))
              ; close to max, set to max and remove force
              (do
;               (print "at max speed")
                (dozer.body:setLinearVelocity max-speed-x max-speed-y)
                (dozer.body:applyForce 0.0 0.0)))))

      ; move backward
      ;
      ; apply a force to come-up to maximum speed
      ;
      ; looks like I have to deal with direction
      ;
      (love.keyboard.isDown "down" "s")

        (if (not (love.keyboard.isDown "up" "w"))
          ; move backward
          (let [max-speed-x (* (math.cos point-dir) (- max-speed))
                max-speed-y (* (math.sin point-dir) (- max-speed))]
;           (print "move backward")
;           (print "max speed x " max-speed-x)
;           (print "max speed y " max-speed-y)
            ; compare magnitude and direction
            ;   backward is large direction difference
            ;
            (if (and (< (+ (sum-of-squares speed-x speed-y) epsilon) dozer.max-speed-2)
                     (> (math.abs (normalize-rad (- point-dir move-dir))) close-dir))
              ; still accelerating
              (let [force-x (- (* (math.cos point-dir)
                                  (/ dozer.accel-force dozer.gear)))
                    force-y (- (* (math.sin point-dir)
                                  (/ dozer.accel-force dozer.gear)))]
;               (print "accelerating")
                (dozer.body:applyForce force-x force-y))
              ; close to max, set to max and remove force
              (do
;               (print "at max speed")
                (dozer.body:setLinearVelocity max-speed-x max-speed-y)
                (dozer.body:applyForce 0.0 0.0)))))

      ; else
      ;
      ; quickly kill any velocity
      ;
      (do
;       (print "no move - kill velocity")
        (if (and (< (math.abs speed-x) epsilon)
                 (< (math.abs speed-y) epsilon))
          ; close to zero, set to zero and remove force
          (do
;           (print "close to zero")
            (dozer.body:setLinearVelocity 0.0 0.0)
            (dozer.body:applyForce 0.0 0.0))
          ; how to distinguish forward from backward
          ;   I do not have to distinguish, the movement direction
          ;   includes this; I just need to reverse the force
          ;
          (let [force-x (- (* (math.cos move-dir)
                              (/ dozer.brake-force dozer.gear)))
                force-y (- (* (math.sin move-dir)
                              (/ dozer.brake-force dozer.gear)))]
            (dozer.body:applyForce force-x force-y))))))

      )

;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;; ;;;;;;;

(fn activate []

  ; start the engine noise
  ;
  (sound.play :running)

  ; initialize the physics world
  ;  uses the STI plugin for Box2D
  ;  I expect this to process the collisions
  ;  for both the tiles and objects
  ;
  (: map :box2d_init world)

  ;; next step?
  ;;
  ;; Set any layer, object, tile, etc to have a custom property called
  ;; "collidable" and give it the value "true". The rocks and bulldozer
  ;; must be collidable, but only some tiles have collision polygons.
  ;;
; (each [key value (pairs map.layers)]
;   (if (= (type key) "string")
;     (set value.collidable true)))

  ; create a custom layer, the map already has a tile layer and an object
  ; layer for rocks, so this is layer 3
  ;
  ;  ***  find a way to get the number of layers from the map
  ;
  ; The alternative is to create the layer with Tiled so I can set
  ; the starting position of the bulldozer.
  ;
  (local player-layer (: map :addCustomLayer "PlayerLayer" 3))

  ; keep it in the game state table
  ;
  (set game-state.player-layer player-layer)

  (set player-layer.collidable true)

  (set player-layer.sprites {
      :dozer {:image dozer.img
              :x dozer.x
              :y dozer.y
              :r dozer.dir } } )

  ; will STI do this for me?
  ;
  ; bulldozer
  ; - make body
  ; - make shape: use a rectangle which covers the image
  ; - make fixture
  ;   newFixture's last arg is density
  ;
  (set dozer.body (love.physics.newBody world dozer.x dozer.y "dynamic"))
  (set dozer.shape (love.physics.newRectangleShape
                     (pixels-to-meters dozer.width)
                     (pixels-to-meters dozer.height)))
  (set dozer.fixture (love.physics.newFixture dozer.body dozer.shape 1))

  ; f = m a
  ;
  ; a three tonne bulldozer will accelerate at 0.5 m/s^2
  ; if a force of 1500 Newtons is applied
  ;
  (dozer.body:setInertia 3000)  ; kilograms
  (dozer.body:setMass    3000)  ; kilograms

  ; I am just pushing the main body without using wheels
  ;
  ; try making two wheels ?
  ; - no need to draw them
  ; - no need for more than one per side
  ;   on the other hand, one at each corner gives resistance to turning
  ; - will I be able to attach them on each side and
  ;   not have to make them collide with rocks & walls
  ;

; ; get the object layer
; ;
; (var object-layer "empty")
; (each [key layer (pairs map.layers)]
;   (if (= layer.name "ObjectLayer1")
;     (set object-layer layer)))

  ; Do I need a draw callback for the objects layer?
  ; No, STI is drawing the rocks.

  ; draw callback for custom layer: the bulldozer
  ;
  ; Will STI draw this for me? Seems it does not draw it.
  ;
  ; draw the bulldozer
  ;
  ; will need to get the position and angle from the physics engine
  ;   body:getPosition returns two values: x, y
  ;
; (love.graphics.setColor 1.0 1.0 1.0 1.0)  ; yes, setColor affects images
; (let [(x y) (dozer.body:getPosition)
;        dir  (dozer.body:getAngle)]
;   (love.graphics.draw dozer.img x y dir
;                       dozer.scale-x dozer.scale-y
;                       dozer.origin-x dozer.origin-y))

  ; from an example
; (fn player-layer.draw []
;   (each [_ sprite (pairs player-layer.sprites)]
;     (let [x (math.floor sprite.x)
;           y (math.floor sprite.y)
;           r sprite.r]
;       (love.graphics.draw sprite.image x y r))))

  (fn player-layer.draw []
    (let [sprite player-layer.sprites.dozer]
      (let [(x y) (dozer.body:getPosition)
            dir  (dozer.body:getAngle)]
        ; assuming the sprite is not automatically updated
        (set sprite.x x)
        (set sprite.y y)
        (set sprite.r dir)
        ; draw the bulldozer
        (love.graphics.draw dozer.img x y dir
                            dozer.scale-x dozer.scale-y
                            dozer.origin-x dozer.origin-y)
        )))

  ; update callback for custom layer
  ;
  (fn player-layer.update [dt]
    (move-bulldozer dt))

  ; dump layers
  ;
; (print "\nlayers")
; (each [key value (pairs map.layers)]
;   (print key value)
;   (each [k v (pairs value)]
;          (print " " k v)))
  )

(fn draw []

  ; draw the map and rocks
  ; the bulldozer is drawn by the call back defined above
  ;
  (: map :draw)
  (: map :box2d_draw 0 0 1 1))

(fn update [dt]

  ; start with the world then the map
  ;
  (: world :update dt)
  (: map :update dt)

  ; normalize the angle of the bulldozer's body
  ;
  (normalize-body-angular dozer.body))


(fn keypressed [key set-state]

  ; shift gears
  ;
  (if (or (= "lshift" key)
          (= "rshift" key))
    (if (= dozer.gear low-gear)
      (set dozer.gear high-gear)
      (set dozer.gear low-gear))))



{:activate activate
 :draw draw
 :update update
 :keypressed keypressed}

